"""BrowserRecorder class for recording snapshots between waits.
"""
import json
import os
import pprint
import time
from contextlib import contextmanager
from enum import Enum
from logging import getLogger
from typing import Optional, List, Tuple, TypeVar

import selenium.webdriver.remote.webdriver
from pydantic import BaseModel, Field
from selenium import webdriver
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

logger = getLogger(__name__)

__all__ = [
    "BrowserRecorder",
    "Waiter",
    "BrowserError",
    "Chrome",
    "Remote",
    "Locator",
    "XPathWithSubstringLocator",
    "SearchMethod",
]


class SearchMethod(Enum):
    """
    An Enum based on selenium's By object, so that values can be explicitly declared.
    """

    ID = By.ID
    XPATH = By.XPATH
    LINK_TEXT = By.LINK_TEXT
    PARTIAL_LINK_TEXT = By.PARTIAL_LINK_TEXT
    NAME = By.NAME
    TAG_NAME = By.TAG_NAME
    CLASS_NAME = By.CLASS_NAME
    CSS_SELECTOR = By.CSS_SELECTOR


class Locator(BaseModel):
    """
    A base class for defining selenium search locators. Essentially this is meant to just be
    passed into selenium's _find_element/_find_elements functions. In some of the high-level selenium abstractions,
    this behavior is invisible, or must be late-bound. This allows for more flexibility with the webdriver_recorder
    API.

    Usage:

        danger_locator = Locator(search_method=SearchMethod.CSS_SELECTOR, search_value='div.panel.panel-danger')
        browser.find_element(*danger_locator.payload)
    """

    search_method: SearchMethod
    search_value: Optional[str]

    @property
    def payload(self) -> Tuple:
        return self.search_method.value, self._search_value

    @property
    def _search_value(self) -> str:
        return self.search_value or ""

    @property
    def state_description(self) -> str:
        return f"wait for {self._search_value} to be visible by method: {self.search_method.value}"


class XPathWithSubstringLocator(Locator):
    """
    The CSS spec does not allow for selectors based on element text, making XPath ideal for
    such searches. This subclass searches for a given tag with the substring displayed; matches are
    case-insensitive.

    locator = XPathWithSubstringLocator(tag='div', displayed_substring='hello')  # will match <div>HELLO</div>
    """

    search_method = Field(SearchMethod.XPATH, const=True)
    tag: str
    displayed_substring: str

    @property
    def _search_value(self) -> str:
        return _xpath_contains(f"//{self.tag}", self.displayed_substring)

    @property
    def state_description(self) -> str:
        return (
            f'wait for {self.tag} with contents "{self.displayed_substring}" '
            f"to be visible via xpath {self._search_value}"
        )


WebDriverType = TypeVar(
    "WebDriverType", bound=selenium.webdriver.remote.webdriver.WebDriver
)


class BrowserRecorder(selenium.webdriver.remote.webdriver.WebDriver):
    """
    A selenium webdriver with extra convenience utilities and
    automatic screenshot capturing.
    """

    pngs: List[bytes] = []  # store screenshots here. intentionally global

    def __init__(self, *args, width=400, height=200, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_window_size(width=width, height=height)
        self.autocapture = True  # automatically capture screenshots

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.quit()

    @contextmanager
    def autocapture_off(self):
        """Context manager temporarily disabling automatic screenshot generation."""
        previous_autocapture = self.autocapture  # for nesting
        self.autocapture = False
        try:
            yield
        finally:
            self.autocapture = previous_autocapture

    def get(self, url):

        return super().get(url)

    def delete_all_cookies(self):
        pass

    def clear(self):
        """Clear the active element."""
        self.switch_to.active_element.clear()

    def click(
        self,
        locator: Locator,
        wait: bool = True,
        timeout: int = 5,
        capture_delay: int = 0,
    ):
        """
        Find tag containing substring and click it.
        wait - give it time to show up in the DOM.
        """
        with self.wrap_exception(locator.state_description):
            if wait and timeout:
                wait = Waiter(self, timeout)
                element = wait.until(
                    EC.element_to_be_clickable(locator.payload),
                    capture_delay=capture_delay,
                )
            else:
                element = self.find_element(*locator.payload)
            element.click()

    def click_tag(self, tag: str, with_substring: str, **kwargs):
        """See wait_for_tag; this does the same thing for clicking on random elements."""
        return self.click(
            XPathWithSubstringLocator(tag=tag, displayed_substring=with_substring),
            **kwargs,
        )

    def click_button(self, substring: str = "", **kwargs):
        """
        Wait for a button with substring to become clickable then click it.
        """
        return self.click_tag("button", substring, **kwargs)

    def wait_for(
        self, locator: Locator, timeout: Optional[int] = None, capture_delay: int = 0
    ):
        """Wait for tag containing substring to show up in the DOM."""
        if timeout is None:
            timeout = getattr(self, "default_wait", 5)
        with self.wrap_exception(locator.state_description):
            wait = Waiter(self, timeout)
            return wait.until(
                EC.visibility_of_element_located(locator.payload),
                capture_delay=capture_delay,
            )

    def wait_for_tag(self, tag: str, with_substring: str, **kwargs):
        """
        Since the XPathWithSubstringLocator is still the best way to search by text, this shim provides
        easy access to it for dependents who want to upgrade to V3.0 but not have to use XPathWithSubstringLocator
        all over the place. This has the same signature as the legacy `wait_for`.
        """
        return self.wait_for(
            XPathWithSubstringLocator(tag=tag, displayed_substring=with_substring),
            **kwargs,
        )

    def run_commands(self, commands):
        """
        Run a series of commands. The first item in a command is the
        browser method to call and the remaining items are the args to
        pass it.
        """
        for method, *args in commands:
            getattr(self, method)(*args)

    def snap(self):
        """
        Store the screenshot as a base64 png in memory.
        Resize the window ahead of time so the full page shows in the shot.
        """
        size = self.get_window_size()
        height = int(self.execute_script("return document.body.scrollHeight"))
        # For remote webdriver the scrollHeight still shows a scroll bar.
        # Bump it up just a little.
        size["height"] = height + 120
        self.set_window_size(**size)
        self.pngs.append(self.get_screenshot_as_base64())

    def send(self, *strings):
        """
        Send the list of strings to the window, with a TAB between each
        string.
        """
        chain = ActionChains(self)
        chain.send_keys(Keys.TAB.join(strings)).perform()

    def send_inputs(self, *strings):
        elements = self.find_elements_by_css_selector("input")
        for element, string in zip(elements, strings):
            element.send_keys(string)

    def hide_inputs(self):
        """Obscure all text inputs on the current screen."""
        javascript = """
            var inputsToHide = document.querySelectorAll('input');
            Array.prototype.forEach.call(inputsToHide, function(el){
                if (!el.attributes['type'] || el.attributes['type'].value === 'text')
                    el.setAttribute('type', 'password');
            })
        """
        self.execute_script(javascript)

    def open_tab(self):
        self.execute_script("window.open('');")
        self.switch_to.window(self.window_handles[-1])

    def close_tab(self):
        self.close()
        if self.window_handles:
            self.switch_to.window(self.window_handles[-1])

    @contextmanager
    def tab_context(self):
        self.open_tab()
        try:
            yield
        finally:
            self.close_tab()

    def send_secret(self, *encrypted_strings):
        """
        Send the list of strings to the window, decrypting them first.
        We do this encrypted to keep the secrets from appearing in
        Tracebacks.
        """
        strings = [self.decrypt(string) for string in encrypted_strings]
        self.send(*strings)

    def decrypt(self, encrypted_text):  # pragma: no cover
        """Method to decrypt text to be overridden."""
        raise NotImplementedError("decrypt should be overridden")

    @contextmanager
    def wrap_exception(self, message):
        """Wrap any exceptions caught in a BrowserError with message."""
        try:
            yield
        except Exception as e:
            if not self.autocapture:
                self.snap()
            raise BrowserError(self, message, str(e))


class Waiter(WebDriverWait):
    """Custom WebDriverWait object that grabs a screenshot after every wait."""

    def __init__(self, driver, timeout, *args, **kwargs):
        super().__init__(driver, timeout, *args, **kwargs)
        self.__driver = driver

    def until(self, *arg, capture_delay: int = 0, **kwargs):
        """
        Every time we wait, take a screenshot of the outcome.
        capture_delay - when we're done waiting, wait just  a little longer
          for whatever animations to take effect.
        """
        try:
            element = super().until(*arg, **kwargs)
        finally:
            if self.__driver.autocapture:
                if capture_delay:
                    time.sleep(capture_delay)
                self.__driver.snap()
        return element


class BrowserError(Exception):
    """Error to raise for a meaningful browser error report."""

    def __init__(self, browser, message, *args):
        self.message = message
        self.url = browser.current_url
        self.logs = browser.get_log("browser")
        self.log_last_http(browser)
        super().__init__(message, self.url, self.logs, *args)

    @staticmethod
    def log_last_http(browser):
        """Log the last http transaction as an error."""

        if "har" not in browser.log_types:
            return
        logs = browser.get_log("har")
        if not logs:
            return
        last_message = json.loads(logs[-1].get("message", {}))
        entries = last_message.get("log", {}).get("entries", [])
        if not entries:
            return
        message = pprint.pformat(entries[-1])
        logger.error(f"Last HTTP transaction: {message}")


def _xpath_contains(node, substring):
    lc_translate = (
        "translate(., 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', 'abcdefghijklmnopqrstuvwxyz')"
    )
    if '"' in substring:
        raise ValueError("double quotes in substring not supported")
    substring = substring.lower()
    return f'{node}[contains({lc_translate}, "{substring}")]'


class Chrome(BrowserRecorder, webdriver.Chrome):
    def __init__(self, *args, options=None, **kwargs):
        if not options:
            options = webdriver.ChromeOptions()
            options.binary_location = os.environ.get("CHROME_BIN")
            options.add_argument('--headless')
        super().__init__(*args, options=options, **kwargs)


class Remote(BrowserRecorder, webdriver.Remote):
    capabilities = DesiredCapabilities



import datetime
import html
import itertools
import json
import logging
import os
import re
import tempfile
from contextlib import contextmanager
from pathlib import Path
from string import ascii_uppercase
from typing import Any, Callable, Dict, List, Optional

import jinja2
import pytest
from _pytest.fixtures import FixtureRequest
from pydantic import BaseModel, root_validator
from selenium import webdriver

from .browser import BrowserError, BrowserRecorder, Chrome, Remote

TEMPLATE_FILE = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "report.template.html"
)

log = logging.getLogger(__name__)

_here = os.path.abspath(os.path.dirname(__file__))


def pytest_addoption(parser):
    group = parser.getgroup("webdriver_recorder")
    group.addoption(
        "--selenium-server",
        action="store",
        dest="selenium_server",
        default=os.environ.get('REMOTE_SELENIUM'),
        help="Remote selenium webdriver to connect to (eg localhost:4444)",
    )
    group.addoption(
        "--report-dir",
        action="store",
        dest="report_dir",
        default=os.environ.get(
            'REPORT_DIR',
            os.path.join(os.getcwd(), "webdriver-report")
        ),
        help="The path to the directory where artifacts should be stored.",
    )
    group.addoption(
        "--jinja-template",
        action="store",
        dest="report_template",
        default=os.path.join(_here, "report.template.html"),
    )
    group.addoption(
        "--report-title",
        action="store",
        dest="report_title",
        default=None,
        help="An optional title for your report; if not provided, "
        "the url of the final test-case executed before generation will be used. "
        "You may provide a constant default by overriding the report_title fixture.",
    )


class HTMLEscapeBaseModel(BaseModel):
    """
    Base model that automatically escapes all strings passed into it.
    """

    @root_validator(pre=True)
    def escape_strings(cls, values) -> Dict[str, Any]:
        values = dict(values)
        for k, v in values.items():
            if isinstance(v, str):
                values[k] = html.escape(str(v))
        return values


class ResultFailure(HTMLEscapeBaseModel):
    message: str
    url: Optional[str]
    loglines: Optional[List[str]]


class ResultAttributes(HTMLEscapeBaseModel):
    link: str
    doc: Optional[str]
    nodeid: str
    pngs: List[bytes]
    failure: Optional[ResultFailure]
    time1: str
    time2: str


class ResultHeader(HTMLEscapeBaseModel):
    link: str
    is_failed: bool
    description: str


class ReportResult(object):
    """
    A test result for passing to the report_test fixture.
    report -- a pytest test outcome
    excinfo -- exception info if there is any
    doc -- the docstring for the test if there is any
    """

    def __init__(self, report, excinfo, doc):
        self.report = report
        self.excinfo = excinfo
        self.doc = doc


@pytest.fixture(scope="session")
def selenium_server(request) -> Optional[str]:
    """Returns a non-empty string or None"""
    return (
        # CLI arg takes precedence
        request.config.getoption("selenium_server")
        # Otherwise, we look for a non-empty string
        or os.environ.get('SELENIUM_SERVER', '').strip()
        # If the result is still Falsey, we always return None.
        or None
    )


@pytest.fixture(scope="session")
def template_filename():
    return TEMPLATE_FILE


@pytest.fixture(scope="session")
def chrome_options():
    options = webdriver.ChromeOptions()
    options.add_argument('--headless')
    options.add_argument("--incognito")
    options.add_argument("--disable-application-cache")
    options.add_argument('--no-sandbox')
    options.add_argument('--disable-dev-shm-usage')
    return options


@pytest.fixture(scope="session")
def session_browser(selenium_server, chrome_options):
    if selenium_server and selenium_server.strip():  # pragma: no cover
        logging.info(f"Creating connection to remote selenium server {selenium_server}")
        browser = Remote(
            options=chrome_options, command_executor=f"http://{selenium_server}/wd/hub"
        )
    else:
        browser = Chrome(options=chrome_options)
    try:
        yield browser
    finally:
        browser.quit()


@pytest.fixture(scope="session")
def browser_context(request: FixtureRequest) -> Callable[..., Chrome]:
    """
    This fixture allows you to create a fresh context for a given
    browser instance.

    The default behavior of the `browser` fixture is to always run in a context of the session scope, so
    you only need to use this if you are not using (or are overriding) the `browser` fixture.

    The fixture itself simply passes the context manager, so you can use it like so:

        def test_something(browser_context):
            with browser_context() as browser:
                browser.get('https://www.uw.edu')

    You may also provide a list of urls to visit to clear cookies at the end of your session,
    if the default 'delete_all_cookies' behavior is not enough to cover your use case.
    """

    @contextmanager
    def inner(
        browser: Optional[Chrome] = None, cookie_urls: Optional[List[str]] = None
    ):
        if not browser:
            # Only loads this fixture if no override is present
            # to avoid creating session_browsers if the
            # dependent does not to.
            browser = request.getfixturevalue('session_browser')
        browser.open_tab()
        cookie_urls = cookie_urls or []
        try:
            yield browser
        finally:
            browser.delete_all_cookies()
            for url in cookie_urls:
                browser.get(url)
                browser.delete_all_cookies()
            browser.close_tab()

    return inner


@pytest.fixture(scope="class")
def class_browser(request, browser_context) -> Chrome:
    with browser_context() as browser:
        request.cls.browser = browser
        yield browser


@pytest.fixture
def browser(browser_context) -> BrowserRecorder:
    """
    The default browser fixture. This default behavior will lazily
    instantiate a `session_browser` if one does not exist. To
    override this behavior and create your own context, you can
    redefine this fixture.
    """
    with browser_context() as browser:
        yield browser


@pytest.fixture(scope="session")
def report_dir(request):
    return request.config.getoption("report_dir")


@pytest.fixture(scope="session", autouse=True)
def report_generator(generate_report, report_dir) -> str:
    """Fixture returning the directory containing our report files. Overridable using `--report-dir`"""
    os.makedirs(report_dir, exist_ok=True)
    worker_file = tempfile.mktemp(prefix="worker.", dir=report_dir)
    Path(worker_file).touch()
    try:
        yield report_dir
    finally:
        os.remove(worker_file)
        workers = list(f for f in os.listdir(report_dir) if f.startswith("worker."))
        if not workers:
            generate_report()


@pytest.fixture(autouse=True)
def report_test(report_generator, request):
    """
    Print the results to report_file after a test run. Without this, the results of the test will not be saved.
    You can ensure this is always run by including the following in your conftest.py:

    @pytest.fixture(autouse=True)
    def report_test(report_test):
        return report_test
    """
    time1 = str(datetime.datetime.now())
    yield
    time2 = str(datetime.datetime.now())
    try:
        # TODO: Sometimes this line causes problems, but I can't
        #       remember the context surrounding it. I think if there
        #       is an error setting up a test fixture, `report_call` is not
        #       defined, or something. Anyway, it'd be a great
        #       to figure out a graceful solution.
        #       In the meantime this'll be nicer output.
        nodeid = request.node.report_call.report.nodeid
    except Exception:  # pragma: no cover
        logging.error(
            f"Unable to extract nodeid from node: {request.node}; "
            "not preparing report segment"
        )
        return
    is_failed = request.node.report_call.report.failed
    # TODO: Figure out a way to include class docs if they exist
    #       class TestFooBar:
    #           """
    #           When Foo is bar
    #           """
    #           def test_a_baz(self):
    #               """and baz is bop"""
    #               do_work('bop')
    #   The report output should then read "When foo is bar and baz is bop"
    doc = request.node.report_call.doc
    slug = re.sub(r"\W", "-", nodeid)
    header = ResultHeader(link=slug, is_failed=is_failed, description=nodeid)
    failure = None
    if is_failed:
        exc_info = request.node.report_call.excinfo
        if isinstance(exc_info.value, BrowserError):
            e = exc_info.value
            failure = ResultFailure(
                message=e.message,
                url=e.url,
                loglines=[log.get("message", "") for log in e.logs],
            )
        else:
            failure = ResultFailure(message=str(exc_info))

    report = ResultAttributes(
        link=slug,
        doc=doc,
        nodeid=nodeid,
        pngs=BrowserRecorder.pngs,
        failure=failure,
        time1=time1,
        time2=time2,
    )

    filename = os.path.join(report_generator, f"result.{slug}.html")
    headerfile = os.path.join(report_generator, f"head.{slug}.html")

    with open(headerfile, "w") as fd:
        fd.write(header.json())
    with open(filename, "w") as fd:
        fd.write(report.json())
    BrowserRecorder.pngs.clear()


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    This gives us hooks from which to report status post test-run.
    Import this into your conftest.py.
    """
    outcome = yield
    report = outcome.get_result()
    if report.when == "call":
        doc = getattr(getattr(item, "function", None), "__doc__", None)
        item.report_call = ReportResult(report=report, excinfo=call.excinfo, doc=doc)


def lettergen():
    """
    Used to generate unique alpha tags.
    :return: A, B, C, ..., AA, AB, AC, ..., BA, BB, BC, ...
    """
    for repeat in range(1, 10):
        for item in itertools.product(ascii_uppercase, repeat=repeat):
            yield "".join(item)


@pytest.fixture(scope="session")
def report_title(request) -> str:
    return request.config.getoption("report_title", default="Webdriver Recorder Summary")


@pytest.fixture(scope="session", autouse=True)
def generate_report(template_filename, report_title, report_dir):
    """
    Uses the included HTML template to generate the final report, using the results found in `report_dir`. Can be
    called explicitly in order to do this at any time.
    """

    def inner(output_dir: Optional[str] = None):
        output_dir = output_dir or report_dir
        with open(template_filename) as fd:
            template = jinja2.Template(fd.read())

        template.globals.update(
            {"date": str(datetime.datetime.now()), "lettergen": lettergen, "zip": zip}
        )

        headers = iterfiles(output_dir, "head.")
        results = iterfiles(output_dir, "result.")
        stream = template.stream(headers=headers, results=results, project=report_title)
        artifact = os.path.join(output_dir, "index.html")
        stream.dump(artifact)
        logging.info(f"Created report: {artifact}")

    return inner


def iterfiles(dir, prefix):
    """
    Iterate through the objects contained in files starting with prefix.
    Delete afterwards.
    """
    files = (f for f in os.listdir(dir) if f.startswith(prefix))
    for filename in sorted(files):
        filename = os.path.join(dir, filename)
        with open(filename) as fd:
            data = json.load(fd)
        yield data
        os.remove(filename)
